/*
 * Handle connections.
 *
 * Copyright (C) 2011  Simon Ruderich
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "tlsproxy.h"
#include "connection.h"
#include "verify.h"

/* close() */
#include <unistd.h>
/* getaddrinfo() */
#include <netdb.h>
/* poll() */
#include <poll.h>
/* errno */
#include <errno.h>


/* Maximum line of a HTTP request line. Longer request lines are aborted with
 * an error. The standard doesn't specify a maximum line length but this
 * should be a good limit to make processing simpler. */
#define MAX_REQUEST_LINE 4096

/* Format string used to send HTTP/1.0 error responses to the client.
 *
 * %s is used 4 times, first three are the error code (no %n$s!), the last is
 * the message. */
#define HTTP_RESPONSE_FORMAT "HTTP/1.0 %s\r\n\
Content-Type: text/html; charset=US-ASCII\r\n\
\r\n\
<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">\n\
<html>\n\
<head><title>%s</title></head>\n\
<body>\n\
<h1>%s</h1>\n\
<p>%s</p>\n\
</body>\n\
</html>\n"


static int initialize_tls_session_client(int peer_socket,
        const char *hostname,
        gnutls_session_t *session,
        gnutls_certificate_credentials_t *x509_cred);
static int initialize_tls_session_server(int peer_socket,
        gnutls_session_t *session,
        gnutls_certificate_credentials_t *x509_cred);

static int read_http_request(FILE *client_fd, char *request, size_t length);
static void send_bad_request(FILE *client_fd);
static void send_forwarding_failure(FILE *client_fd);
static void tls_send_invalid_cert_message(gnutls_session_t session);

static void transfer_data(int client, int server);
static int read_from_write_to(int from, int to);
static void transfer_data_tls(int client, int server,
                              gnutls_session_t client_session,
                              gnutls_session_t server_session);
static int read_from_write_to_tls(gnutls_session_t from, gnutls_session_t to);

static int connect_to_host(const char *hostname, const char *port);

static int parse_request(const char *buffer, char *host, char *port,
                                             int *version_minor);


void handle_connection(int client_socket) {
    int server_socket;
    FILE *client_fd, *server_fd;

    char buffer[MAX_REQUEST_LINE];
    char host[MAX_REQUEST_LINE];
    char port[5 + 1];

    int version_minor; /* HTTP/1.x */
    int result;

    /* client_x509_cred is used when talking to the client (acting as a TSL
     * server), server_x509_cred is used when talking to the server (acting as
     * a TSL client). */
    gnutls_certificate_credentials_t client_x509_cred, server_x509_cred;

    gnutls_session_t client_session, server_session;
    /* initialize_tls_session_*() called? - used for goto out */
    int client_session_init, server_session_init;
    /* gnutls_handshake() called? - used for goto out */
    int client_session_started, server_session_started;
    /* Validation failed? If yes we need to send the special "invalid"
     * certificate. */
    int validation_failed;

    LOG(LOG_DEBUG, "new connection");

    server_socket = -1;
    client_fd = NULL;
    server_fd = NULL;
    client_session_init = 0;
    server_session_init = 0;
    client_session_started = 0;
    server_session_started = 0;
    validation_failed = 0;

    client_fd = fdopen(client_socket, "a+");
    if (NULL == client_fd) {
        LOG_PERROR(LOG_WARNING, "fdopen(): client failed");
        goto out;
    }

    /* Read request line (CONNECT ..) and headers (they are discarded). */
    result = read_http_request(client_fd, buffer, sizeof(buffer));
    if (-1 == result) {
        /* Read error. */
        LOG(LOG_WARNING, "read_http_request(): client read error");
        goto out;
    } else if (-2 == result) {
        /* EOF */
        LOG(LOG_WARNING, "read_http_request(): client EOF");
        send_bad_request(client_fd);
        goto out;
    }

    if (0 != parse_request(buffer, host, port, &version_minor)) {
        LOG(LOG_WARNING, "bad request: %s", buffer);
        send_bad_request(client_fd);
        goto out;
    }

    LOG(LOG_DEBUG, "target: %s:%s (HTTP 1.%d)", host, port, version_minor);

    /* Connect to proxy server or directly to server. */
    if (NULL != global_proxy_host && NULL != global_proxy_port) {
        LOG(LOG_DEBUG, "connecting to %s:%s", global_proxy_host,
                                              global_proxy_port);
        server_socket = connect_to_host(global_proxy_host, global_proxy_port);
    } else {
        LOG(LOG_DEBUG, "connecting to %s:%s", host, port);
        server_socket = connect_to_host(host, port);
    }

    if (-1 == server_socket) {
        LOG(LOG_WARNING, "failed to connect to server");
        send_forwarding_failure(client_fd);
        goto out;
    }
    server_fd = fdopen(server_socket, "a+");
    if (NULL == server_fd) {
        LOG_PERROR(LOG_WARNING, "fdopen(): server failed");
        send_forwarding_failure(client_fd);
        goto out;
    }

    /* Connect to proxy if requested (command line option). */
    if (NULL != global_proxy_host && NULL != global_proxy_port) {
        fprintf(server_fd, "CONNECT %s:%s HTTP/1.0\r\n", host, port);
        fprintf(server_fd, "\r\n");

        /* Read response line from proxy server. */
        result = read_http_request(server_fd, buffer, sizeof(buffer));
        if (-1 == result) {
            /* Read error. */
            LOG(LOG_WARNING, "read_http_request(): proxy read error");
            send_forwarding_failure(client_fd);
            goto out;
        } else if (-2 == result) {
            /* EOF */
            LOG(LOG_WARNING, "read_http_request(): proxy EOF");
            send_forwarding_failure(client_fd);
            goto out;
        }

        /* Check response of proxy server. */
        if (0 != strncmp(buffer, "HTTP/1.0 200", 12)) {
            LOG(LOG_WARNING, "bad proxy response: %s", buffer);
            send_forwarding_failure(client_fd);
            goto out;
        }
    }

    LOG(LOG_DEBUG, "connection to server established");

    /* If the -u option is used and we don't know this hostname's server
     * certificate then just pass through the connection and let the client
     * verify the server certificate. */
    if (global_passthrough_unknown) {
        char path[1024];
        FILE *file = NULL;

        if (-2 == server_certificate_path(&file, host, path, sizeof(path))) {
            /* We've established a connection, tell the client. */
            fprintf(client_fd, "HTTP/1.0 200 Connection established\r\n");
            fprintf(client_fd, "\r\n");
            fflush(client_fd);

            LOG(LOG_DEBUG, "transferring data");

            /* Proxy data between client and server until one side is done
             * (EOF or error). */
            transfer_data(client_socket, server_socket);

            LOG(LOG_DEBUG, "finished transferring data");

            goto out;
        }
        /* server_certificate_path() may open the file, close it. */
        if (NULL != file) {
            fclose(file);
        }
    }

    /* Initialize TLS client credentials to talk to the server. */
    result = initialize_tls_session_server(server_socket, &server_session,
                                                          &server_x509_cred);
    if (0 != result) {
        LOG(LOG_WARNING, "initialize_tls_session_server() failed");
        send_forwarding_failure(client_fd);
        goto out;
    }
    server_session_init = 1;

    LOG(LOG_DEBUG, "starting server TLS handshake");

    /* Try to establish TLS handshake between us and server. */
    result = gnutls_handshake(server_session);
    if (GNUTLS_E_SUCCESS != result) {
        LOG(LOG_WARNING, "server TLS handshake failed: %s",
                         gnutls_strerror(result));
        send_forwarding_failure(client_fd);
        goto out;
    }
    server_session_started = 1;

    LOG(LOG_DEBUG, "server TLS handshake finished");

    /* Make sure the server certificate is valid and known. */
    if (0 != verify_tls_connection(server_session, host)) {
        LOG(LOG_ERROR, "server certificate validation failed!");
        /* We send the error message over our TLS connection to the client,
         * but with an invalid certificate. No data is transfered from/to the
         * target server. */
        validation_failed = 1;
    }

    /* Initialize TLS server credentials to talk to the client. */
    result = initialize_tls_session_client(client_socket,
                                           /* use special host if the server
                                            * certificate was invalid */
                                           (validation_failed) ? "invalid"
                                                               : host,
                                           &client_session,
                                           &client_x509_cred);
    if (0 != result) {
        LOG(LOG_WARNING, "initialize_tls_session_client() failed");
        send_forwarding_failure(client_fd);
        goto out;
    }
    client_session_init = 1;

    /* We've established a connection, tell the client. */
    fprintf(client_fd, "HTTP/1.0 200 Connection established\r\n");
    fprintf(client_fd, "\r\n");
    fflush(client_fd);

    LOG(LOG_DEBUG, "starting client TLS handshake");

    /* Try to establish TLS handshake between client and us. */
    result = gnutls_handshake(client_session);
    if (GNUTLS_E_SUCCESS != result) {
        LOG(LOG_WARNING, "client TLS handshake failed: %s",
                         gnutls_strerror(result));
        send_forwarding_failure(client_fd);
        goto out;
    }
    client_session_started = 1;

    LOG(LOG_DEBUG, "client TLS handshake finished");

    /* Tell the client that the verification failed. Shouldn't be necessary as
     * the client should terminate the connection because he received the
     * invalid certificate but better be sure. */
    if (validation_failed) {
        tls_send_invalid_cert_message(client_session);
        goto out;
    }

    LOG(LOG_DEBUG, "transferring TLS data");

    /* Proxy data between client and server until one side is done (EOF or
     * error). */
    transfer_data_tls(client_socket, server_socket,
                      client_session, server_session);

    LOG(LOG_DEBUG, "finished transferring TLS data");

out:
    /* Close TLS sessions if necessary. Use GNUTLS_SHUT_RDWR so the data is
     * reliable transmitted. */
    if (0 != server_session_started) {
        gnutls_bye(server_session, GNUTLS_SHUT_RDWR);
    }
    if (0 != client_session_started) {
        gnutls_bye(client_session, GNUTLS_SHUT_RDWR);
    }
    if (0 != server_session_init) {
        gnutls_deinit(server_session);
        gnutls_certificate_free_credentials(server_x509_cred);
    }
    if (0 != client_session_init) {
        gnutls_deinit(client_session);
        gnutls_certificate_free_cas(client_x509_cred);
        gnutls_certificate_free_keys(client_x509_cred);
        gnutls_certificate_free_credentials(client_x509_cred);
    }

    /* Close connection to server/proxy. */
    if (NULL != server_fd) {
        fclose(server_fd);
    } else if (-1 != server_socket) {
        close(server_socket);
    }
    LOG(LOG_DEBUG, "connection to server closed");
    /* Close connection to client. */
    if (NULL != client_fd) {
        fclose(client_fd);
    } else {
        close(client_socket);
    }
    LOG(LOG_DEBUG, "connection to client closed");

    LOG(LOG_DEBUG, "connection finished");
}


static int initialize_tls_session_client(int peer_socket,
        const char *hostname,
        gnutls_session_t *session,
        gnutls_certificate_credentials_t *x509_cred) {
    int result;
    int use_invalid_cert;
    char path[1024];

    /* The "invalid" hostname is special. If it's used we send an invalid
     * certificate to let the client know something is wrong. */
    use_invalid_cert = 0 == strcmp(hostname, "invalid");

    /* Hostname too long. */
    if (sizeof(path) - strlen(PROXY_SERVER_CERT_FORMAT) <= strlen(hostname)) {
        LOG(LOG_WARNING,
            "initialize_tls_session_client(): hostname too long: '%s'",
            hostname);
        return -1;
    }
    /* Try to prevent path traversals in hostnames. */
    if (NULL != strstr(hostname, "..")) {
        LOG(LOG_WARNING,
            "initialize_tls_session_client(): possible path traversal: '%s'",
            hostname);
        return -1;
    }
    snprintf(path, sizeof(path), PROXY_SERVER_CERT_FORMAT, hostname);

    result = gnutls_certificate_allocate_credentials(x509_cred);
    if (GNUTLS_E_SUCCESS != result) {
        LOG(LOG_ERROR,
            "initialize_tls_session_client(): \
gnutls_certificate_allocate_credentials(): %s",
            gnutls_strerror(result));
        return -1;
    }

    /* Load proxy CA file, this CA "list" is send to the client. */
    if (!use_invalid_cert) {
        result = gnutls_certificate_set_x509_trust_file(*x509_cred,
                                                        PROXY_CA_FILE,
                                                        GNUTLS_X509_FMT_PEM);
        if (0 >= result) {
            LOG(LOG_ERROR,
                "initialize_tls_session_client(): can't read CA file: '%s'",
                PROXY_CA_FILE);
            gnutls_certificate_free_credentials(*x509_cred);
            return -1;
        }
    }
    /* If the invalid hostname was specified do nothing, we use a self-signed
     * certificate in this case. */

    /* And certificate for this website and proxy's private key. */
    if (!use_invalid_cert) {
        result = gnutls_certificate_set_x509_key_file(*x509_cred,
                                                      path, PROXY_KEY_FILE,
                                                      GNUTLS_X509_FMT_PEM);
    /* If the invalid hostname was specified load our special "invalid"
     * certificate. */
    } else {
        result = gnutls_certificate_set_x509_key_file(*x509_cred,
                                                      PROXY_INVALID_CERT_FILE,
                                                      PROXY_KEY_FILE,
                                                      GNUTLS_X509_FMT_PEM);
    }
    if (GNUTLS_E_SUCCESS != result) {
        LOG(LOG_ERROR,
            "initialize_tls_session_client(): \
can't read server certificate ('%s') or key file ('%s'): %s",
            path, PROXY_KEY_FILE, gnutls_strerror(result));
        gnutls_certificate_free_credentials(*x509_cred);
        /* Could be a missing certificate. */
        return -2;
    }

    gnutls_certificate_set_dh_params(*x509_cred, global_tls_dh_params);

    result = gnutls_init(session, GNUTLS_SERVER);
    if (GNUTLS_E_SUCCESS != result) {
        LOG(LOG_ERROR,
            "initialize_tls_session_client(): gnutls_init(): %s",
            gnutls_strerror(result));
        gnutls_certificate_free_credentials(*x509_cred);
        return -1;
    }
    result = gnutls_priority_set(*session, global_tls_priority_cache);
    if (GNUTLS_E_SUCCESS != result) {
        LOG(LOG_ERROR,
            "initialize_tls_session_client(): gnutls_priority_set(): %s",
            gnutls_strerror(result));
        gnutls_deinit(*session);
        gnutls_certificate_free_credentials(*x509_cred);
        return -1;
    }
    result = gnutls_credentials_set(*session,
                                    GNUTLS_CRD_CERTIFICATE, *x509_cred);
    if (GNUTLS_E_SUCCESS != result) {
        LOG(LOG_ERROR,
            "initialize_tls_session_client(): gnutls_credentials_set(): %s",
            gnutls_strerror(result));
        gnutls_deinit(*session);
        gnutls_certificate_free_credentials(*x509_cred);
        return -1;
    }

    gnutls_transport_set_ptr(*session, (gnutls_transport_ptr_t)peer_socket);

    return 0;
}
static int initialize_tls_session_server(int peer_socket,
        gnutls_session_t *session,
        gnutls_certificate_credentials_t *x509_cred) {
    int result;

    result = gnutls_certificate_allocate_credentials(x509_cred);
    if (GNUTLS_E_SUCCESS != result) {
        LOG(LOG_ERROR,
            "initialize_tls_session_server(): \
gnutls_certificate_allocate_credentials(): %s",
            gnutls_strerror(result));
        return -1;
    }

    result = gnutls_init(session, GNUTLS_CLIENT);
    if (GNUTLS_E_SUCCESS != result) {
        LOG(LOG_ERROR,
            "initialize_tls_session_server(): gnutls_init(): %s",
            gnutls_strerror(result));
        gnutls_certificate_free_credentials(*x509_cred);
        return -1;
    }
    gnutls_priority_set(*session, global_tls_priority_cache);
    if (GNUTLS_E_SUCCESS != result) {
        LOG(LOG_ERROR,
            "initialize_tls_session_server(): gnutls_priority_set(): %s",
            gnutls_strerror(result));
        gnutls_deinit(*session);
        gnutls_certificate_free_credentials(*x509_cred);
        return -1;
    }
    result = gnutls_credentials_set(*session,
                                    GNUTLS_CRD_CERTIFICATE, *x509_cred);
    if (GNUTLS_E_SUCCESS != result) {
        LOG(LOG_ERROR,
            "initialize_tls_session_server(): gnutls_credentials_set(): %s",
            gnutls_strerror(result));
        gnutls_deinit(*session);
        gnutls_certificate_free_credentials(*x509_cred);
        return -1;
    }

    gnutls_transport_set_ptr(*session, (gnutls_transport_ptr_t)peer_socket);

    return 0;
}


/* Read HTTP request line and headers (ignored).
 *
 * On success 0 is returned, -1 on client error, -2 on unexpected EOF.
 */
static int read_http_request(FILE *client_fd, char *request, size_t length) {
    char buffer[MAX_REQUEST_LINE];

    if (NULL == fgets(request, (int)length, client_fd)) {
        if (ferror(client_fd)) {
            LOG_PERROR(LOG_WARNING, "read_http_request(): fgets()");
            return -1;
        }

        return -2;
    }

    while (NULL != fgets(buffer, MAX_REQUEST_LINE, client_fd)) {
        /* End of header. */
        if (0 == strcmp(buffer, "\n") || 0 == strcmp(buffer, "\r\n")) {
            break;
        }
    }
    if (ferror(client_fd)) {
        LOG_PERROR(LOG_WARNING, "read_http_request(): fgets()");
        return -1;
    }

    return 0;
}

static void send_bad_request(FILE *client_fd) {
#define RESPONSE_ERROR "400 Bad Request"
#define RESPONSE_MSG   "Your browser sent an invalid request."
    fprintf(client_fd, HTTP_RESPONSE_FORMAT,
                       RESPONSE_ERROR, RESPONSE_ERROR, RESPONSE_ERROR,
                       RESPONSE_MSG);
#undef RESPONSE_ERROR
#undef RESPONSE_MSG
}
static void send_forwarding_failure(FILE *client_fd) {
#define RESPONSE_ERROR "503 Forwarding failure"
#define RESPONSE_MSG   "Failed to connect to server, check logs."
    fprintf(client_fd, HTTP_RESPONSE_FORMAT,
                       RESPONSE_ERROR, RESPONSE_ERROR, RESPONSE_ERROR,
                       RESPONSE_MSG);
#undef RESPONSE_ERROR
#undef RESPONSE_MSG
}
static void tls_send_invalid_cert_message(gnutls_session_t session) {
#define RESPONSE_ERROR "500 Internal Server Error"
#define RESPONSE_MSG   "Server certificate validation failed, check logs."

    char buffer[sizeof(HTTP_RESPONSE_FORMAT) - 1 /* '\0' */
                                             - 4 * 2 /* four %s */
                + (sizeof(RESPONSE_ERROR) - 1 /* '\0' */) * 3
                + sizeof(RESPONSE_MSG)    - 1 /* '\0' */
                + 1 /* '\0' */];

    snprintf(buffer, sizeof(buffer),
             HTTP_RESPONSE_FORMAT,
             RESPONSE_ERROR, RESPONSE_ERROR, RESPONSE_ERROR, RESPONSE_MSG);

    gnutls_record_send(session, buffer, sizeof(buffer) - 1);
                                        /* don't send trailing '\0' */
#undef RESPONSE_ERROR
#undef RESPONSE_MSG
}


/* Transfer data between client and server sockets until one closes the
 * connection. */
static void transfer_data(int client, int server) {
    struct pollfd fds[2];
    fds[0].fd      = client;
    fds[0].events  = POLLIN | POLLPRI | POLLHUP | POLLERR;
    fds[0].revents = 0;
    fds[1].fd      = server;
    fds[1].events  = POLLIN | POLLPRI | POLLHUP | POLLERR;
    fds[1].revents = 0;

    for (;;) {
        int result = poll(fds, 2, -1 /* no timeout */);
        if (result < 0) {
            LOG_PERROR(LOG_ERROR, "transfer_data(): poll()");
            return;
        }

        /* Data available from client. */
        if (fds[0].revents & POLLIN || fds[0].revents & POLLPRI) {
            if (0 != read_from_write_to(client, server)) {
                /* EOF (or other error) */
                break;
            }
        }
        /* Data available from server. */
        if (fds[1].revents & POLLIN || fds[1].revents & POLLPRI) {
            if (0 != read_from_write_to(server, client)) {
                /* EOF (or other error) */
                break;
            }
        }

        /* Client closed connection. */
        if (fds[0].revents & POLLERR || fds[0].revents & POLLHUP) {
            break;
        }
        /* Server closed connection. */
        if (fds[1].revents & POLLERR || fds[1].revents & POLLHUP) {
            break;
        }
    }
}

/* Read available data from socket from and write it to socket to. At maximum
 * 4096 bytes are read/written. */
static int read_from_write_to(int from, int to) {
    ssize_t size_read;
    ssize_t size_written;
    char buffer[4096];

    LOG(LOG_DEBUG, "read_from_write_to(): %d -> %d", from, to);

    size_read = read(from, buffer, sizeof(buffer));
    if (0 > size_read) {
        LOG_PERROR(LOG_WARNING, "read_from_write_to(): read()");
        return -1;
    }
    /* EOF */
    if (0 == size_read) {
        return -1;
    }

    size_written = write(to, buffer, (size_t)size_read);
    if (0 > size_written) {
        LOG_PERROR(LOG_WARNING, "read_from_write_to(): write()");
        return -1;
    }
    if (size_read != size_written) {
        LOG(LOG_ERROR, "read_from_write_to(): only written %ld of %ld bytes!",
                       (long int)size_written, (long int)size_read);
        return -1;
    }

    return 0;
}

/* Transfer data between client and server TLS connection until one closes the
 * connection. */
static void transfer_data_tls(int client, int server,
                              gnutls_session_t client_session,
                              gnutls_session_t server_session) {
    struct pollfd fds[2];
    fds[0].fd      = client;
    fds[0].events  = POLLIN | POLLPRI | POLLHUP | POLLERR;
    fds[0].revents = 0;
    fds[1].fd      = server;
    fds[1].events  = POLLIN | POLLPRI | POLLHUP | POLLERR;
    fds[1].revents = 0;

    for (;;) {
        int result = poll(fds, 2, -1 /* no timeout */);
        if (result < 0) {
            LOG_PERROR(LOG_ERROR, "transfer_data(): poll()");
            return;
        }

        /* Data available from client. */
        if (fds[0].revents & POLLIN || fds[0].revents & POLLPRI) {
            if (0 != read_from_write_to_tls(client_session, server_session)) {
                /* EOF (or other error) */
                break;
            }
        }
        /* Data available from server. */
        if (fds[1].revents & POLLIN || fds[1].revents & POLLPRI) {
            if (0 != read_from_write_to_tls(server_session, client_session)) {
                /* EOF (or other error) */
                break;
            }
        }

        /* Client closed connection. */
        if (fds[0].revents & POLLERR || fds[0].revents & POLLHUP) {
            break;
        }
        /* Server closed connection. */
        if (fds[1].revents & POLLERR || fds[1].revents & POLLHUP) {
            break;
        }
    }
}

/* Read available data from session from and write to session to. */
static int read_from_write_to_tls(gnutls_session_t from,
                                  gnutls_session_t to) {
    size_t size;
    ssize_t size_read;
    ssize_t size_written;
    char buffer[16384];

    /* Get maximum possible buffer size. */
    size = gnutls_record_get_max_size(from);
    LOG(LOG_DEBUG, "read_from_write_to_tls(): suggested buffer size: %ld",
                   (long int)size);
    if (size > gnutls_record_get_max_size(to)) {
        size = gnutls_record_get_max_size(to);
    }
    if (size > sizeof(buffer)) {
        size = sizeof(buffer);
    }
    LOG(LOG_DEBUG, "read_from_write_to_tls(): used buffer size: %ld",
                   (long int)size);

    size_read = gnutls_record_recv(from, buffer, size);
    if (0 > size_read) {
        LOG(LOG_WARNING, "read_from_write_to_tls(): gnutls_record_recv(): %s",
                         gnutls_strerror((int)size_read));
        return -1;
    }
    /* EOF */
    if (0 == size_read) {
        return -1;
    }

    size_written = gnutls_record_send(to, buffer, (size_t)size_read);
    if (0 > size_written) {
        LOG(LOG_WARNING, "read_from_write_to_tls(): gnutls_record_send(): %s",
                         gnutls_strerror((int)size_written));
        return -1;
    }
    if (size_read != size_written) {
        LOG(LOG_ERROR, "read_from_write_to_tls(): only written %ld of %ld bytes!",
                       (long int)size_written, (long int)size_read);
        return -1;
    }

    return 0;
}


static int connect_to_host(const char *hostname, const char *port) {
    struct addrinfo gai_hints;
    struct addrinfo *gai_result;
    int gai_return;

    int server_socket;
    struct addrinfo *server;

    if (NULL == hostname || NULL == port) {
        return -1;
    }

    /* Get IP of hostname server. */
    memset(&gai_hints, 0, sizeof(gai_hints));
    gai_hints.ai_family   = AF_UNSPEC;
    gai_hints.ai_socktype = SOCK_STREAM;
    gai_hints.ai_protocol = 0;
    gai_hints.ai_flags    = AI_NUMERICSERV /* given port is numeric */
                          | AI_ADDRCONFIG  /* supported by this computer */
                          | AI_V4MAPPED;   /* support IPv4 through IPv6 */
    gai_return = getaddrinfo(hostname, port, &gai_hints, &gai_result);
    if (0 != gai_return) {
        LOG_PERROR(LOG_WARNING, "connect_to_host(): getaddrinfo()");
        return -1;
    }

    /* Now try to connect to each server returned by getaddrinfo(), use the
     * first successful connect. */
    for (server = gai_result; NULL != server; server = server->ai_next) {
        server_socket = socket(server->ai_family,
                               server->ai_socktype,
                               server->ai_protocol);
        if (-1 == server_socket) {
            LOG_PERROR(LOG_DEBUG, "connect_to_host(): socket(), trying next");
            continue;
        }

        if (-1 != connect(server_socket, server->ai_addr,
                                         server->ai_addrlen)) {
            break;
        }
        LOG_PERROR(LOG_DEBUG, "connect_to_host(): connect(), trying next");

        close(server_socket);
    }
    /* Make sure we free the result from getaddrinfo(). */
    freeaddrinfo(gai_result);

    if (NULL == server) {
        LOG_PERROR(LOG_WARNING, "connect_to_host(): no server found, abort");
        return -1;
    }

    return server_socket;
}


/* Parse HTTP CONNECT request string and save its parameters.
 *
 * The following format is expected: "CONNECT host:port HTTP/1.x".
 *
 * request and host must have the same size! port must be at least 6 bytes
 * long (5 + '\0').
 */
static int parse_request(const char *request, char *host, char *port,
                                              int *version_minor) {
    int port_unused; /* just used to verify the port is numeric */
    char *position;

    /* scanf() doesn't check spaces. */
    if (0 != strncmp(request, "CONNECT ", 8)) {
        return -1;
    }
    /* Check request and extract data, "host:port" is not yet separated. */
    if (2 != sscanf(request, "CONNECT %s HTTP/1.%d",
                             host, version_minor)) {
        return -1;
    }
    /* Make sure ":port" is there. */
    if (NULL == (position = strchr(host, ':'))) {
        return -1;
    }
    /* Make sure port is numeric. */
    if (1 != sscanf(position + 1, "%d", &port_unused)) {
        return -1;
    }
    /* Store it in *port. */
    strncpy(port, position + 1, 5);
    port[5] = '\0';
    /* And remove port from host. */
    *position = '\0';

    return 0;
}
